#include <glib.h>
#include <mysql.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>

#include <xchat/xchat-plugin.h>

static xchat_plugin *ph;
MYSQL *mysql=NULL;

#define MYSQL_HOST "localhost"
#define MYSQL_DB   "uno"
#define MYSQL_USER "xuno"
#define MYSQL_PASS "xuno"

#define MAX_CARDS 64
#define DECK_SIZE 108
#define MAX_CARD_NAME 128
#define START_CARDS 7

static gchar *
_mysql_escape_string (MYSQL   *mysql,
                      const gchar *str)
{
  gint cachelen;
  gchar *retval;

  cachelen = strlen ( str );
  retval = g_malloc ( cachelen*2 + 1 );

  mysql_real_escape_string ( mysql, retval, str, cachelen );
  return retval;
}

static MYSQL_RES*
_mysql_query (MYSQL    *mysql,
              const gchar *query,
              gint         len)
{
	MYSQL_RES *retval;

	if (mysql_real_query(mysql, query, len))
	{
		xchat_printf(ph, "mysql_real_query failed: %s\n", mysql_error(mysql));
		return NULL;
	}

	if (!(retval=mysql_store_result(mysql)))
	{
		if (!mysql_errno(mysql))
			return NULL;
		xchat_printf(ph, "mysql_store_result failed: %s\n", mysql_error(mysql) );
		return NULL;
	}

	return retval;
}

static MYSQL_RES *
mysql_query_va (MYSQL *mysql,
                const gchar *fmt, /* supports %d and %s */
                ...)
{
	va_list va;
	gchar *query;
	gint qs=0;
	MYSQL_RES *retval;

	query=g_malloc(qs+1);
	*query=0;
	va_start (va, fmt);
	while (*fmt)
	{
		if ((*fmt)=='%')
		{
			fmt++;
			switch (*fmt)
			{
				case 's':
					{
						gchar *s = va_arg(va, char *);
						if (!s)
						{
							qs+=4;
							query=g_realloc(query, qs+1);
							strcat(query, "NULL");
						} else {
							gchar *temp = _mysql_escape_string ( mysql, s );
							qs+=strlen(temp)+2;
							query=g_realloc(query, qs+1);
							strcat(query, "\"");
							strcat(query, temp);
							strcat(query, "\"");
							g_free(temp);
						}
						break;
					}
				case 'd':
					{
						gchar s[32];
						gint d = va_arg(va, int);
						g_snprintf(s, 32, "%d", d);
						qs+=strlen(s);
						query=g_realloc(query, qs+1);
						strcat(query, s);
						break;
					}
				case '%':
					{
						qs++;
						query=g_realloc(query, qs+1);
						query[qs-1]=*fmt;
						query[qs]=0;
						break;
					}
			}
			fmt++;
		} else {
			qs++;
			query=g_realloc(query, qs+1);
			query[qs-1]=*fmt++;
			query[qs]=0;
		}
	}
	va_end (va);
	retval = _mysql_query( mysql, query, qs );
	g_free(query);
	return retval;
}

struct card_t {
	char color;
	char value;
};

struct player_t {
	char *nick; /* empty nick meens local user, and space meens BOT */
	char *hostmask;
	
	struct card_t cards[MAX_CARDS];
	int cards_n;
	
	struct player_t *prev, *next;
	int bot;
};

struct game_t {
	xchat_context *context;
	char *channel; /* strdup */
	char *network; /* strdup of the xchat_get_info(ph (context), "network") if set */

	struct card_t deck[DECK_SIZE];
	int cards_n;

	struct card_t card;

	int draw;
	int status; /* 0 = not active
		       1 = previous game just done, annoytimer points for this to end
		       2 = waiting for players to join, annoytimer points to select done
		       3 = waiting for player to play card
		     */

	struct player_t *players, *player;
	int players_n;
	int direction; /* 0 next, 1 prev */
	xchat_hook *annoytimer;
	time_t starttime;
	int played;
	int bots;
	struct game_t *next, *prev;
};

struct game_t *root=NULL;

static int status1_callback(void *userdata);
static int status2_callback(void *userdata);
static int status3_callback(void *userdata);
static int status20_21_callback(void *userdata);

static int dump_cards(struct game_t *game, struct player_t *player);
static void get_card(struct game_t *game, struct card_t *dst);
static void select_player(struct game_t *game, struct player_t *player); /* sets up 2 minute timer aswell */
static struct player_t *get_next_player(struct game_t *game);
static void player_cards(struct game_t *game, struct player_t *player); /* cards */
static void game_stop(struct game_t *game, struct player_t *player); /* !stop, player can be NULL */
static void player_draw(struct game_t *game, struct player_t *player); /* draw */
static void player_play(struct game_t *game, struct player_t *player, const char *line); /* play */
static void player_color(struct game_t *game, struct player_t *player, const char *line); /* color */
static void player_pass(struct game_t *game, struct player_t *player); /* pass */
static void bot_player(struct game_t *game, struct player_t *player);
static void game_card(struct game_t *game); /* card */
static struct player_t *game_join(struct game_t *game, const char *mask);
static void game_players(struct game_t *game);
static void game_xuno(struct game_t *game); /* !xuno */

static void player_uno(struct game_t *game, struct player_t *player);

static struct player_t *get_next_player(struct game_t *game)
{
	if (game->status<=2)
		return game->players;
	if (game->direction)
	{
		struct player_t *curr;
		if (game->player->prev)
			return game->player->prev;
		curr=game->players;
		while (curr->next)
			curr=curr->next;
		return curr;
	} else {
		if (game->player->next)
			return game->player->next;
		return game->players;
	}
}

static const char *split_mask(const char *mask, char **hostmask)
{
	static char nick[128];
	*hostmask=strchr(mask, '!');
	strncpy(nick, mask, sizeof(nick)-1);
	nick[sizeof(nick)-1]=0;
	if (*hostmask)
	{
		nick[*hostmask-mask]=0;
		(*hostmask)++;
	} else
		*hostmask="";
	return nick;
}

static struct player_t *find_player(struct game_t *game, const char *mask)
{
	char *hostmask;
	const char *nick=split_mask(mask, &hostmask);

	struct player_t *player=game->players;

	while (player)
	{
		if (!strcasecmp(player->nick, nick))
		{
			if (!strcasecmp(hostmask, player->hostmask))
				return player;
		}
		player=player->next;
	}
	player=game->players;
	while (player)
	{
		if (!strcasecmp(player->nick, nick))
		{
			free(player->hostmask);
			player->hostmask=strdup(hostmask);
			return player;
		} else if (!strcasecmp(hostmask, player->hostmask))
		{
			free(player->nick);
			player->nick=strdup(nick);
			return player;
		}
		player=player->next;
	}
	return NULL;
}

static void bot_play_color(struct game_t *game, struct player_t *player, struct card_t *card)
{
	char line[5];
	strcpy(line, "co ");
	switch (card->color)
	{
		default:
		case 1: strcat(line, "r"); break;
		case 2: strcat(line, "g"); break;
		case 3: strcat(line, "b"); break;
		case 4: strcat(line, "y"); break;
	}
	player_color(game, player, line);
}

static void bot_play_card(struct game_t *game, struct player_t *player, struct card_t *card)
{
	char line[10]; /* 9 should do */
	strcpy(line, "pl ");
	switch (card->color)
	{
		case 1: strcat(line, "r "); break;
		case 2: strcat(line, "g "); break;
		case 3: strcat(line, "b "); break;
		case 4: strcat(line, "y "); break;
	}
	switch (card->value)
	{
		case 0:  strcat(line, "0"); break;
		case 1:  strcat(line, "1"); break;
		case 2:  strcat(line, "2"); break;
		case 3:  strcat(line, "3"); break;
		case 4:  strcat(line, "4"); break;
		case 5:  strcat(line, "5"); break;
		case 6:  strcat(line, "6"); break;
		case 7:  strcat(line, "7"); break;
		case 8:  strcat(line, "8"); break;
		case 9:  strcat(line, "9"); break;
		case 10: strcat(line, "s"); break;
		case 11: strcat(line, "d t"); break;
		case 12: strcat(line, "r"); break;
		case 20: strcat(line, "w"); break;
		case 21: strcat(line, "w d f"); break;
	}
	player_play(game, player, line);
}

static void bot_player(struct game_t *game, struct player_t *player)
{
	int i;
head:
	for (i=0;i<player->cards_n;i++)
		if (player->cards[i].color==game->card.color)
		{
			bot_play_card(game, player, player->cards+i);
			goto done;
		}
	for (i=0;i<player->cards_n;i++)
		if ((player->cards[i].value==game->card.value)&&(game->card.value<20))
		{
			bot_play_card(game, player, player->cards+i);
			goto done;
		}
	for (i=0;i<player->cards_n;i++)
		if ((player->cards[i].value>=20))
		{
			bot_play_card(game, player, player->cards+i);
			for (i=0;i<player->cards_n;i++) /* if we won, we don't have any cards to check */
				if (player->cards[i].color)
				{
					bot_play_color(game, player, player->cards+i);
					goto done;
				}
			if (player->cards_n)
				bot_play_color(game, player, &game->card);
			goto done;
		}
	if (!game->draw)
	{
		player_draw(game, player);
		goto head;
	}
	player_pass(game, player);
done:
	if (game->player->cards_n)
		if (game->player->bot)
			bot_player(game, game->player);
}

static struct player_t *game_join(struct game_t *game, const char *mask)
{
	char *hostmask;
	const char *nick=split_mask(mask, &hostmask);

	struct player_t *player=malloc(sizeof(struct player_t)), *last=game->players;

	int i;

	if (game->status<2)
	{
		xchat_commandf(ph, "MSG %s \002\0034No game is currently running (yet?)", game->channel);
		return NULL;
	}

	player->next=NULL;
	if (last)
	{
		while (last->next)
			last=last->next;
		last->next=player;
		player->prev=last;
	} else {
		player->prev=NULL;
		game->players=player;
	}
	game->players_n++;
	
	player->hostmask=strdup(hostmask);
	player->nick=strdup(nick);
	player->cards_n=START_CARDS;
	player->bot=!strcmp(player->hostmask, "~root@localhost");

	for (i=0;i<START_CARDS;i++)
		get_card(game, player->cards+i);
	if (!player->bot)
		player_cards(game, player);
	xchat_commandf(ph, "MSG %s %s \0037joins \002\0034x\0036U\0032N\0033O\0038!", game->channel, player->nick);
	return player;
}

static void get_card(struct game_t *game, struct card_t *dst)
{
	if (!game->cards_n)
	{
		struct card_t cards[DECK_SIZE];
		int i, j, k;
		i=0;
		for (k=0;k<4;k++)
		{
			cards[i].color=0;
			cards[i].value=20; /* wild */
			i++;
		}
		for (k=0;k<4;k++)
		{
			cards[i].color=0;
			cards[i].value=21; /* wild draw four */
			i++;
		}
		for (j=1;j<=4;j++)
		{
			for (k=0;k<=9;k++)
			{
				cards[i].color=j;
				cards[i].value=k;
				i++;
			}
			for (k=1;k<=9;k++)
			{
				cards[i].color=j;
				cards[i].value=k;
				i++;
			}
			for (k=0;k<2;k++)
			{
				cards[i].color=j;
				cards[i].value=10; /* skip */
				i++;
			}
			for (k=0;k<2;k++)
			{
				cards[i].color=j;
				cards[i].value=11; /* draw two*/
				i++;
			}
			for (k=0;k<2;k++)
			{
				cards[i].color=j;
				cards[i].value=12; /* reverse */
				i++;
			}
		}
		for (i=DECK_SIZE;i;i--)
		{
			rand(); rand(); rand();
			k=(int)(((float)i)*rand()/(RAND_MAX+1.0));
			memcpy(game->deck+game->cards_n, cards+k, sizeof(struct card_t));
			game->cards_n++;
			memcpy(cards+k, cards+k+1, (i-k-1)*sizeof(struct card_t));
		}
	}
	memcpy(dst, game->deck+game->cards_n-1, sizeof(struct card_t));
	game->cards_n--;
}

static struct game_t *find_game(const char *channel)
{ /* used by all management functions */
	xchat_context *context=xchat_get_context(ph);
	struct game_t *game=root;
	while (game)
	{
		if (game->context==context) /* is this enough ? */
			if (!strcmp(channel, game->channel))
				return game;
		game=game->next;
	}
	return NULL;
}

static void resolve_card(struct card_t *card, char *dst)
{
	if (card->color)
	{
		switch (card->color)
		{
			default:
			case 1: strcpy(dst, "\0030,4 Red"); break; /* r */
			case 2: strcpy(dst, "\0030,3 Green"); break; /* g */
			case 3: strcpy(dst, "\0030,12 Blue"); break; /* b */
			case 4: strcpy(dst, "\0031,8 Yellow"); break; /* y */
		}
		strcat(dst, " ");
		switch (card->value)
		{
			default:
				{
					char buffer[3];
					buffer[0]=card->value+'0';
					buffer[1]=' ';
					buffer[2]=0;
					strcat(dst, buffer);
					break;
				}
			case 10:
				strcat(dst, "\002Skip ");
				break;
			case 11:
				strcat(dst, "\002Draw Two ");
				break;
			case 12:
				strcat(dst, "\002Reverse ");
				break;
			case 13: /* just color is selected */
				break;
		}
		strcat(dst, "\017");
	} else {
		if (card->value==20)
			strcpy(dst, "\0030,4 W\0030,3I\0030,12L\0031,8D \017");
		else
			strcpy(dst, "\0030,4 W\0030,3I\0030,12L\0031,8D\0030,4 \0030,3D\0030,12r\0031,8a\0030,4w\0030,3 \0030,12F\0031,8o\0030,4u\0030,3r \017");
	}	
}

static void player_uno(struct game_t *game, struct player_t *player)
{
	if (player->cards_n==1)
		xchat_commandf(ph, "MSG %s \0037Player\017 %s \0037has \002\0034x\0036U\0032N\0033O\0038!", game->channel, player->nick);
}

static int dump_cards(struct game_t *game, struct player_t *player)
{
	int retval=0;
	char buffer[512];
	char temp[MAX_CARD_NAME];
	int used=0;
	int i=0;
	if (!player->cards_n)
		return 0;
	buffer[0]=0;
	while (i<player->cards_n)
	{
		if (player->cards[i].value<10)
			retval+=player->cards[i].value;
		else if (player->cards[i].color)
			retval+=20;
		else
			retval+=50;
		resolve_card(player->cards+i, temp);
		if ((strlen(buffer)+strlen(temp)+10)>512)
		{
			xchat_command(ph, buffer);
			used=0;
		}
		if (!used)
		{
			sprintf(buffer, "MSG %s %s:", game->channel, player->nick);
			used=1;
		}
		strcat(buffer, "  ");
		strcat(buffer, temp);
		i++;
	}
	xchat_command(ph, buffer);
	return retval;
}

static void player_cards(struct game_t *game, struct player_t *player) /* cards */
{
	char buffer[512];
	char temp[MAX_CARD_NAME];
	int used=0;
	int i=0;
	if (player->bot)
		return;
	while (i<player->cards_n)
	{
		resolve_card(player->cards+i, temp);
		if ((strlen(buffer)+strlen(temp)+10)>512)
		{
			xchat_command(ph, buffer);
			used=0;
		}
		if (!used)
		{
			sprintf(buffer, "NOTICE %s Hand:", player->nick);
			used=1;
		}
		strcat(buffer, "  ");
		strcat(buffer, temp);
		i++;
	}
	if (used)
		xchat_command(ph, buffer);
}

static void game_stop(struct game_t *game, struct player_t *player) /* !stop, player can be NULL */
{
	struct player_t *curr, *next;
	if (!game->status)
		return;
	if (!player)
	{
		if (game->status>2)
			return;
		if (game->status==2)
			xchat_commandf(ph, "MSG %s \002\0034x\0036U\0032N\0033O\0038!\017 \00323,17stopped", game->channel);
	} else {
		xchat_commandf(ph, "MSG %s \002\0034x\0036U\0032N\0033O\0038!\017 \0037stopped by \0035[\00311%s!%s\0035]", game->channel, player->nick, player->hostmask);
	}

	game->status=0;
	curr=game->players;
	while (curr)
	{
		next=curr->next;
		free(curr->nick);
		free(curr->hostmask);
		free(curr);
		curr=next;
	}
	game->players=NULL;
	game->players_n=0;
	if (game->annoytimer)
	{
		xchat_unhook(ph, game->annoytimer);
		game->annoytimer=0;
	}
}

static void player_draw(struct game_t *game, struct player_t *player) /* draw */
{
	char temp[MAX_CARD_NAME];

	if (game->status!=3)
		return;
	if (game->player!=player)
	{
		xchat_commandf(ph, "NOTICE %s \002\0034You're not up", player->nick);
		return;
	}
	if (game->draw)
	{
		xchat_commandf(ph, "NOTICE %s \002\0034You have already drawn a card", player->nick);
		return;
	}
	xchat_commandf(ph, "MSG %s \0037Player\017 %s \0037Drew a card", game->channel, player->nick);
	game->draw=1;
	if ((player->cards_n+2)>=MAX_CARDS) /* too many cards on hand */
		return;
	get_card(game, player->cards+player->cards_n);
	resolve_card(player->cards+player->cards_n, temp);
	player->cards_n++;
	if (!player->bot)
		xchat_commandf(ph, "NOTICE %s \0037Draw card: %s", player->nick, temp);
	if (game->annoytimer)
		xchat_unhook(ph, game->annoytimer);
	game->annoytimer=xchat_hook_timer(ph, 120000, status3_callback, game);
}

static void player_pass(struct game_t *game, struct player_t *player) /* pass */
{
	char temp[MAX_CARD_NAME];
	struct player_t *next=get_next_player(game);
	if (game->status!=3)
		return;
	if (game->player!=player)
	{
		xchat_commandf(ph, "NOTICE %s \002\0034You're not up", player->nick);
		return;
	}
	if (!game->draw)
	{
		xchat_commandf(ph, "NOTICE %s \002\0034You haven't drawn a card yet", player->nick);
		return;
	}
	resolve_card(&game->card, temp);
	xchat_commandf(ph, "MSG %s \0037Player\017 %s \0037passes over to \017%s \0037(Top card is %s\0037)", game->channel, player->nick, next->nick, temp);
	select_player(game, next);
	if (game->player->bot)
		bot_player(game, game->player);
	else
		player_cards(game, game->player);
}

static void player_play(struct game_t *game, struct player_t *player, const char *line) /* play */
{
	char temp[MAX_CARD_NAME];
	char temp1[MAX_CARD_NAME];
	char temp2[MAX_CARD_NAME];
	char temp3[MAX_CARD_NAME];
	char temp4[MAX_CARD_NAME];
	const char *tmp;
	struct card_t card, card1, card2, card3, card4;
	struct player_t *next, *next2;
	int i;
	if (game->status!=3)
		return;
	if (game->player!=player)
	{
		xchat_commandf(ph, "NOTICE %s \002\0034You're not up", player->nick);
		return;
	}
	if (!(tmp=strchr(line, ' ')))
		goto invalid_card;
	tmp++;
	if ((!strcasecmp(tmp, "w")))
	{
		card.color=0;
		card.value=20;
		goto got_value;
	}
	if ((!strcasecmp(tmp, "w d f")))
	{
		card.color=0;
		card.value=21;
		goto got_value;
	}
	if ((!strncasecmp(tmp, "r ", 2)))
	{
		card.color=1;
		tmp+=2;
		goto got_color;
	}
	if ((!strncasecmp(tmp, "g ", 2)))
	{
		card.color=2;
		tmp+=2;
		goto got_color;
	}
	if ((!strncasecmp(tmp, "b ", 2)))
	{
		card.color=3;
		tmp+=2;
		goto got_color;
	}
	if ((!strncasecmp(tmp, "y ", 2)))
	{
		card.color=4;
		tmp+=2;
		goto got_color;
	}
	goto invalid_card;
got_color:
	if ((!strcasecmp(tmp, "r")))
	{
		card.value=12;
		goto got_value;
	}
	if ((!strcasecmp(tmp, "s")))
	{
		card.value=10;
		goto got_value;
	}
	if ((!strcasecmp(tmp, "d t")))
	{
		card.value=11;
		goto got_value;
	}
	if ((strlen(tmp)==1)&&(isdigit(*tmp)))
	{
		card.value=*tmp-'0';
		goto got_value;
	}
	goto invalid_card;
got_value:
	for (i=0;i<player->cards_n;i++)
	{
		if ((player->cards[i].value==card.value)&&(player->cards[i].color==card.color))
			break;
	}
	if (i==player->cards_n)
	{
		if (!player->bot)
			xchat_commandf(ph, "NOTICE %s \002\0034You don't have that card", player->nick);
		return;
	}
	if  ( ! ( (!card.color) || ((game->card.color==card.color) || (card.value==game->card.value) ) ) )
	{
		if (!player->bot)
			xchat_commandf(ph, "NOTICE %s \002\0034You can't use this card", player->nick);
		return;
	}

	memmove(player->cards+i, player->cards+i+1, sizeof(struct card_t)*(player->cards_n-i-1));
	player->cards_n--;
	game->played++;
	resolve_card(&card, temp);

	/* Premature this message before any drew messages */
	if (!player->cards_n)
		xchat_commandf(ph, "MSG %s \0037Player\017 %s \0037plays %s \0037and wins!", game->channel, player->nick, temp);

	switch (card.value)
	{
		case 11: /* draw two */
			next2=get_next_player(game);
			get_card(game, &card1);
			get_card(game, &card2);
			resolve_card(&card1, temp1);
			resolve_card(&card2, temp2);
			if (!player->cards_n)
				if (!next2->bot)
					xchat_commandf(ph, "NOTICE %s You got cards:  %s  %s", next2->nick, temp1, temp2);
			if ((next2->cards_n+2)<MAX_CARDS)
			{
				memcpy(next2->cards+next2->cards_n, &card1, sizeof(card1));
				next2->cards_n++;
			}
			if ((next2->cards_n+2)<MAX_CARDS)
			{
				memcpy(next2->cards+next2->cards_n, &card2, sizeof(card2));
				next2->cards_n++;
			}
			break;
		case 21:
			next2=get_next_player(game);
			get_card(game, &card1);
			get_card(game, &card2);
			get_card(game, &card3);
			get_card(game, &card4);
			resolve_card(&card1, temp1);
			resolve_card(&card2, temp2);
			resolve_card(&card3, temp3);
			resolve_card(&card4, temp4);
			if (!player->cards_n)
				if (!next2->bot)
					xchat_commandf(ph, "NOTICE %s You got cards:  %s  %s  %s  %s", next2->nick, temp1, temp2, temp3, temp4);
			if ((next2->cards_n+2)<MAX_CARDS)
			{
				memcpy(next2->cards+next2->cards_n, &card1, sizeof(card1));
				next2->cards_n++;
			}
			if ((next2->cards_n+2)<MAX_CARDS)
			{
				memcpy(next2->cards+next2->cards_n, &card2, sizeof(card2));
				next2->cards_n++;
			}
			if ((next2->cards_n+2)<MAX_CARDS)
			{
				memcpy(next2->cards+next2->cards_n, &card3, sizeof(card3));
				next2->cards_n++;
			}
			if ((next2->cards_n+2)<MAX_CARDS)
			{
				memcpy(next2->cards+next2->cards_n, &card4, sizeof(card4));
				next2->cards_n++;
			}
			break;
	}
	
	if (!player->cards_n)
	{
		struct player_t *iter=game->players;;
		int prawn=0;
		time_t usedtime=time(NULL)-game->starttime;
		char nick[256];
		MYSQL_RES *res;
		MYSQL_ROW row;
		while (iter)
		{
			if (iter!=player)
				prawn+=dump_cards(game, iter);
			iter=iter->next;
		}
		xchat_commandf(ph, "MSG %s Prawns: %d. Cards played: %d. Time: %d minutes and %d seconds!", game->channel, prawn, game->played, usedtime/60, usedtime%60);
		if ((res=mysql_query_va(mysql, "SELECT nick from players where nick LIKE %s", player->nick)))
		{
			if ((row=mysql_fetch_row(res)))
			{
				strncpy(nick, row[0], 255);
				nick[255]=0;
				goto nick_done;
			}
			mysql_free_result(res);
		}
		strcpy(nick, player->nick);
		mysql_query_va(mysql, "INSERT INTO players (nick, color) VALUES (%s, %d)", nick, 19+(int) (11.0*rand()/(RAND_MAX+1.0)));
nick_done:
		mysql_query_va(mysql, "INSERT INTO game (nick, prawn, cardsplayed, timeused, timestamp) VALUES (%s, %d, %d, %d, UNIX_TIMESTAMP())", nick, prawn, game->played, usedtime);

		game->status=1;
		game_stop(game, NULL);
		game->status=1;
		xchat_commandf(ph, "MSG %s \002\0034New game will start in 30 seconds", game->channel);
		if (game->annoytimer)
			xchat_unhook(ph, game->annoytimer);
		game->annoytimer=xchat_hook_timer(ph, 30000, status1_callback, game);
		return;
	}

	switch (card.value)
	{
		case 10: /* skip */
			next2=get_next_player(game);
			select_player(game, next2);
			/* give cards to this person */
			next=get_next_player(game);
			xchat_commandf(ph, "MSG %s %s \0037plays %s \0037on\017 %s\0037 and next player up is\017 %s", game->channel, player->nick, temp, next2->nick, next->nick);
			break;	
		case 11: /* draw two */
/*			next2=get_next_player(game);*/
			select_player(game, next2);
			next=get_next_player(game);
			xchat_commandf(ph, "MSG %s %s \0037plays %s\0037 on\017 %s \0037and next player up is\017 %s", game->channel, player->nick, temp, next2->nick, next->nick);
			if (!next2->bot)
				xchat_commandf(ph, "NOTICE %s You got cards:  %s  %s", next2->nick, temp1, temp2);
			break;

		case 12: /* reverse */
			game->direction^=1;
			if (game->players_n!=2)
				next=get_next_player(game);
			else
				next=player;
			xchat_commandf(ph, "MSG %s %s \0037plays\017 %s\0037, and next player up is\017 %s", game->channel, player->nick, temp, next->nick);
			break;
		case 20:
			next=get_next_player(game);
			game->status=20;
			xchat_commandf(ph, "MSG %s %s \0037plays %s \0037to\017 %s", game->channel, player->nick, temp, next->nick);
			player_uno(game, player);
			xchat_commandf(ph, "MSG %s %s\0037: Please select color", game->channel, player->nick);
			if (game->annoytimer)
				xchat_unhook(ph, game->annoytimer);
			game->annoytimer=xchat_hook_timer(ph, 120000, status20_21_callback, game);
			return;
		case 21:
			next=get_next_player(game);
			xchat_commandf(ph, "MSG %s %s \0037plays %s \0037on\017 %s\0037, who now will be skipped", game->channel, player->nick, temp, next->nick);
			if (!next2->bot)
				xchat_commandf(ph, "NOTICE %s You got cards:  %s  %s  %s  %s", next2->nick, temp1, temp2, temp3, temp4);
			player_uno(game, player);
			xchat_commandf(ph, "MSG %s %s\0037: Please select color", game->channel, player->nick);
			game->status=21;
			if (game->annoytimer)
				xchat_unhook(ph, game->annoytimer);
			game->annoytimer=xchat_hook_timer(ph, 120000, status20_21_callback, game);
			return;
		default:
			next=get_next_player(game);
			xchat_commandf(ph, "MSG %s %s \0037plays %s\0037, and next player up is\017 %s", game->channel, player->nick, temp, next->nick);
			break;
	}
	player_uno(game, player);
	memcpy(&game->card, &card, sizeof(card));
	select_player(game, next);
	if (game->player->bot)
		bot_player(game, game->player);
	else
		player_cards(game, game->player);
	
	 /* TODO */
	 /* todo play card */
	 /* todo did we win */
	 /* wodo select color?, timer */
	 /* else todo select player */
	 /* timer timer timer */
	return;

invalid_card:
	if (!player->bot)
		xchat_commandf(ph, "NOTICE %s \002\0034Invalid card", player->nick);
	return;
}

static void game_card(struct game_t *game) /* !CARD */
{
	char temp[MAX_CARD_NAME];
	if (game->status<=1)
		return;
	resolve_card(&game->card, temp);
	xchat_commandf(ph, "MSG %s \0037Card is: %s", game->channel, temp);
}

static void game_players(struct game_t *game)
{
	struct player_t *player=game->players;
	int count=1;
	while (player)
	{
		int tag=0;
		if (game->status>2)
			if (game->player==player)
				tag=1;
		if (tag)
			xchat_commandf(ph, "MSG %s \0037Player \00309%d\0037 is \0035[\00311%s!%s\0035] (active player)", game->channel, count, player->nick, player->hostmask);
		else
			xchat_commandf(ph, "MSG %s \0037Player \00309%d\0037 is \0035[\00311%s!%s\0035]", game->channel, count, player->nick, player->hostmask);
		count++;
		player=player->next;
	}
}

static void select_player(struct game_t *game, struct player_t *player) /* sets up 2 minute timer aswell */
{
	if (game->annoytimer)
		xchat_unhook(ph, game->annoytimer);
	game->player=player;
	game->status=3;
	game->draw=0;
	game->annoytimer=xchat_hook_timer(ph, 120000, status3_callback, game);
}

static void player_color(struct game_t *game, struct player_t *player, const char *line) /* color */
{
	char temp[MAX_CARD_NAME];
	const char *tmp;
	struct card_t card;
	struct player_t *was=game->player, *next = get_next_player(game);

	if ((game->status!=21)&&(game->status!=20))
		return;
	if (game->player!=player)
	{
		if (!game->player->bot)	
			xchat_commandf(ph, "NOTICE %s \002\0034You're not up", player->nick);
		return;
	}
	if (!(tmp=strchr(line, ' ')))
		goto invalid_color;
	tmp++;
	card.value=13;
	if (!strcasecmp(tmp, "r"))
		card.color=1;
	else if (!strcasecmp(tmp, "g"))
		card.color=2;
	else if (!strcasecmp(tmp, "b"))
		card.color=3;
	else if (!strcasecmp(tmp, "y"))
		card.color=4;
	else goto invalid_color;
	resolve_card(&card, temp);
	if (game->status==21)
	{
		select_player(game, next);
		next=get_next_player(game);
	}
	xchat_commandf(ph, "MSG %s \0037Player\017 %s \0037selects color %s \0037and play continues with\017 %s", game->channel, was->nick, temp, next->nick);
	memcpy(&game->card, &card, sizeof(card));
	select_player(game, next);
	if (game->player->bot)
		bot_player(game, game->player);
	else
		player_cards(game, game->player);
	return;
invalid_color:
	if (!game->player->bot)	
		xchat_commandf(ph, "NOTICE %s \002\0034Invalid color", player->nick);
	return;
}

static int status20_21_callback(void *userdata)
{
	char temp[MAX_CARD_NAME];
	struct game_t *game=userdata;
	struct player_t *was=game->player, *next = get_next_player(game);
	struct card_t card;

	game->annoytimer=0;

	card.color=game->card.color;
	card.value=13;

	resolve_card(&card, temp);
	memcpy(&game->card, &card, sizeof(card));
	xchat_set_context(ph, game->context);
	if (game->status==21)
	{
		select_player(game, next);
		next=get_next_player(game);
	}
	xchat_commandf(ph, "MSG %s \0037Player\017 %s \0037was idle two minutes. Selecting color %s \0037and game continues with\017 %s", game->channel, was->nick, temp, next->nick);
	select_player(game, next);
	if (!game->player->bot) /* in case we have a bug */
		player_cards(game, game->player);
	return 0;
}

static int status3_callback(void *userdata)
{
	struct game_t *game=userdata;
	struct player_t *next = get_next_player(game);

	xchat_set_context(ph, game->context);
	game->annoytimer=0;

	xchat_commandf(ph, "MSG %s %s \0037Has been idle 2 minutes, skipping to player\017 %s", game->channel, game->player->nick, next->nick);
	select_player(game, next);
	if (!game->player->bot) /* in case we have a bug */
		player_cards(game, game->player);
	else
		bot_player(game, game->player);
	return 0; /* kill timer */
}

static void game_ai(struct game_t *game)
{
	switch(game->bots++)
	{
		case 0: game_join(game, "Mother!~root@localhost"); break;
		case 1: game_join(game, "Father!~root@localhost"); break;
		case 2: game_join(game, "Bishop!~root@localhost"); break;
		case 3: game_join(game, "Alice!~root@localhost"); break;
		case 4: game_join(game, "R2D2!~root@localhost"); break;
		default:
			xchat_commandf(ph, "MSG %s \002\0034Can't create more AI's now", game->channel);
			game->bots--;
	}
}

static int status2_callback(void *userdata)
{
	struct game_t *game=userdata;
	char temp[MAX_CARD_NAME];

	xchat_set_context(ph, game->context);
	game->annoytimer=0;

	if (!game->players)
	{
		xchat_commandf(ph, "MSG %s \002\0034No players, disabling game", game->channel);
		game_stop(game, NULL);
		return 0;
	} else if (game->players_n==1)
		game_ai(game);

	game->direction=0;
	do {
		get_card(game, &game->card);
	} while (!game->card.color);
	resolve_card(&game->card, temp);

	select_player(game, game->players);
	xchat_commandf(ph, "MSG %s %s \0037Plays First - The Top Card is: %s", game->channel, game->player->nick, temp);
	game->starttime=time(NULL);
	game->played=0;

	return 0; /* kill timer */
}

static int status1_callback(void *userdata)
{
	struct game_t *game=userdata;

	xchat_set_context(ph, game->context);
	game->annoytimer=0;

	game_xuno(game);
	
	return 0; /* kill timer */
}

static void game_xuno(struct game_t *game)
{
	if (game->status>=2)
	{
		xchat_commandf(ph, "MSG %s \002\0034Game already running", game->channel);
	} else {
		xchat_commandf(ph, "MSG %s \002\0034x\0036U\0032N\0033O\0038!\017 \0031,6v0.0.7 by mw|", game->channel);
		xchat_commandf(ph, "MSG %s \002\0034x\0036U\0032N\0033O\0038!\017 \00316,12Join Now - You Have 30 Seconds ", game->channel);
		game->status=2;
		if (game->annoytimer)
			xchat_unhook(ph, game->annoytimer);
		game->annoytimer=xchat_hook_timer(ph, 30000, status2_callback, game);
		game->bots=0;
	}
	return;
}

/*
 * mw|!~sortre@208.80-203-0.nextgentel.com    <- mask format
 */
static void parse(struct game_t *game, const char *mask, const char *line)
{
	struct player_t *player = find_player(game, mask);

	if (!strcasecmp(line, "!xuno"))
	{
		game_xuno(game);
		return;
	}
	
	if (player&&(!strcasecmp(line, "cards")))
	{
		player_cards(game, player);
		return;
	}

	if (!player&&(!strcasecmp(line, "join")))
	{
		/*player=*/game_join(game, mask);
		return;
	}
	if (player&&((!strcasecmp(line, "ai"))||(!strcasecmp(line, "!ai"))))
	{
		game_ai(game);
		return;
	}
	if ((!strcasecmp(line, "!stop")))
	{
		game_stop(game, player);
		return;
	}
	if ((!strcasecmp(line, "card")))
	{
		game_card(game);
		return;
	}
	if ((!strcasecmp(line, "!players")))
	{
		game_players(game);
		return;
	}
	if (player&&((!strncasecmp(line, "pl ", 3))||(!strncasecmp(line, "play ", 5))))
	{
		player_play(game, player, line);
		return;
	}
	if (player&&((!strncasecmp(line, "co ", 3))||(!strncasecmp(line, "color ", 6))))
	{
		player_color(game, player, line);
		return;
	}
	if (player&&((!strcasecmp(line, "pa"))||(!strcasecmp(line, "pass"))))
	{
		player_pass(game, player);
		return;
	}
	if (player&&((!strcasecmp(line, "dr"))||(!strcasecmp(line, "draw"))))
	{
		player_draw(game, player);
		return;
	}

}

static int message_PRIVMSG(char *word[], char *word_eol[], void *userdata) /* inbound message */
{
	struct game_t *game;

	if ((game=find_game(word[3])))
	{
		parse(game, word[1]+1, word_eol[4]+1);
		return XCHAT_EAT_NONE; /* :-p */
	}
	return XCHAT_EAT_NONE;
}
static int message_USERMSG(char *word[], char *word_eol[], void *userdata) /* outbound message */
{
	struct game_t *game;

	if (!(game=find_game(xchat_get_info(ph, "channel"))))
		return XCHAT_EAT_NONE;
	parse(game, xchat_get_info(ph, "nick"), word_eol[1]);
	return XCHAT_EAT_NONE;
}

static int uno_hook(char *word[], char *word_eol[], void *userdata)
{
	struct game_t *game;
	const char *channel = xchat_get_info(ph, "channel");

	if (find_game(channel))
	{
		xchat_printf(ph, "\002\0034Game already created for this channel\n");
		return XCHAT_EAT_ALL;
	}


	if (!channel)
	{
not_channel:
		xchat_printf(ph, "\002\0034You are not in a channel\n");
		return XCHAT_EAT_ALL;
	}
	if (channel[0]!='#')
		goto not_channel;

	xchat_printf(ph, "\002\0034Creating a game for %s\n", channel);

	game=malloc(sizeof(struct game_t));
	game->next=root;
	game->prev=NULL;
	if (root)
		root->prev=game;
	root=game;

	game->context=xchat_get_context(ph);
	game->channel=strdup(channel);
	if ((game->network=(char *)xchat_get_info(ph, "network")))
		game->network=strdup(game->network);

	game->cards_n=0;
	game->status=0;
	game->annoytimer=0;
	game->players=0;
	game->players_n=0;

	return XCHAT_EAT_ALL;
}

static int uno_unhook(char *word[], char *word_eol[], void *userdata)
{
	struct game_t *game = find_game(xchat_get_info(ph, "channel"));
	if (!game)
	{
		xchat_printf(ph, "\002\0034Game not found\n");
		return XCHAT_EAT_ALL;	
	}
	
	game_stop(game, NULL);
	
	free(game->channel);
	if (game->network)
		free(game->network);	
	if (game->prev)
		game->prev->next=game->next;
	else
		root=game->next;
	if (game->next)
		game->next->prev=game->prev;
	free(game);
	xchat_printf(ph, "\002\0034Game destroyed\n");
	return XCHAT_EAT_ALL;
}

int xchat_plugin_init(xchat_plugin *plugin_handle,
                      char **plugin_name,
                      char **plugin_desc,
                      char **plugin_version,
                      char *arg)
{
	/* we need to save this for use with any xchat_* functions */
	ph = plugin_handle;
	*plugin_name = "UNO";
	*plugin_desc = "We like to entertain";
	*plugin_version = "0.1";

	xchat_print(ph, "Attemping to init UNO plugin");

	if (!(mysql=mysql_init(NULL)))
		return 0;
	if (!mysql_real_connect(mysql, MYSQL_HOST, MYSQL_USER, MYSQL_PASS, MYSQL_DB, MYSQL_PORT, NULL, 0))
	{
		xchat_printf(ph, "Failed to connect to database: Error: %s\n", mysql_error(mysql));
		return 0;
	}

	xchat_hook_command(ph, "uno_hook", XCHAT_PRI_NORM, uno_hook, "Usage: uno_hook", 0);
	xchat_hook_command(ph, "uno_unhook", XCHAT_PRI_NORM, uno_unhook, "Usage: uno_unhook", 0);

	xchat_hook_command(ph, "", XCHAT_PRI_NORM, message_USERMSG, NULL, 0);
	xchat_hook_server(ph, "PRIVMSG", XCHAT_PRI_NORM, message_PRIVMSG, 0);

	return 1; /* return 1 for success */
}

__attribute__((destructor)) static void lib_close(void)
{
	struct game_t *game=root, *next;
	while (game)
	{
		next=game->next;
		if (game->status>=2)
		{
			struct player_t *player, *nextp;
			player=game->players;
			while (player)
			{
				nextp=player->next;
				free(player->nick);
				free(player->hostmask);
				free(player);
				player=nextp;
			}
		}
		free(game->channel);
		if (game->network)
			free(game->network);	
		free(game);
		game=next;
	}
	if (mysql)
		mysql_close(mysql);
}
